/**
 * Extend the base Actor entity by defining a custom roll data structure which is ideal for the Simple system.
 * @extends {Actor}
 */

import { RdDUtility } from "./rdd-utility.js";
import { RdDRollDialog } from "./rdd-roll-dialog.js";

export class RdDActor extends Actor {

  /* -------------------------------------------- */
  /**
   * Override the create() function to provide additional RdD functionality.
   *
   * This overrided create() function adds initial items 
   * Namely: Basic skills, money, 
   *
   * @param {Object} data        Barebones actor data which this function adds onto.
   * @param {Object} options     (Unused) Additional options which customize the creation workflow.
   *
   */
   
  static async create(data, options) {
    // If the created actor has items (only applicable to duplicated actors) bypass the new actor creation logic
    if (data.items) {
      return super.create(data, options);
    }
    super.create(data, options);
    
  }
  
  /* -------------------------------------------- */
  
  prepareData() {
    super.prepareData();

    const actorData = this.data;
    const data = actorData.data;
    const flags = actorData.flags;

    // Make separate methods for each Actor type (character, npc, etc.) to keep
    // things organized.
    if (actorData.type === 'personnage') this._prepareCharacterData(actorData);
  }

  /* -------------------------------------------- */
  /**
   * Prepare Character type specific data
   */
  _prepareCharacterData(actorData) {
    // Initialize empty items
    RdDUtility.computeCarac(actorData.data);
    this.computeEtatGeneral();
  }
  
  /* -------------------------------------------- */
  async performRoll( rollData ) {
    
    let myroll = new Roll("d100");
    myroll.roll();
    let quality = "Echec";
    let xpmsg = "";
    let tache = 0;
    //console.log(">>> ROLL", rollData.selectedCarac.label, rollData.rollTarget.score, myroll.total );
    let result =  myroll.total;
    if (result <= rollData.rollTarget.part) {
      quality = "Réussite Particulière!";
      if ( rollData.finalLevel < 0 ) {
        let xpcarac = Math.floor( Math.abs(rollData.finalLevel) / 2);
        let xpcomp  = (Math.abs(rollData.finalLevel) % 2 == 1) ? xpcarac+1 : xpcarac;
        xpmsg = "<br>Points d'expérience gagné ! " + xpcarac + " - " + xpcomp;
      }
      rollData.pointsDeTache = 4;  
      rollData.qualite = 2;
    } else if (result <= (rollData.rollTarget.score /2) ) {
      quality = "Réussite Significative";
      rollData.pointsDeTache = 2;
      rollData.qualite = 1;
    } else if (result <= (rollData.rollTarget.score) ) {
      quality = "Réussite Normale";
      rollData.pointsDeTache = 1;
      rollData.qualite = 0;
    } else if (result < (rollData.rollTarget.epart) ) {
      quality = "Echec Normal";
      rollData.pointsDeTache = 0;
      rollData.qualite = -2;
    } else if (result < (rollData.rollTarget.etotal) ) {
      quality = "Echec Particulier";
      rollData.pointsDeTache = -2;      
      rollData.qualite = -4;
    } else if (result >= (rollData.rollTarget.etotal) ) {
      quality = "Echec Total";
      rollData.pointsDeTache = -4;
      rollData.qualite = -6;
    }
    
    // Fight management !
    let defenseMsg;
    let specialStr = "<br>Points de taches : " + rollData.pointsDeTache; // Per default
    if ( rollData.arme ) { // In case of fight, replace the "tache" per dommages + localization. "tache" indicates if result is OK or not
      if ( rollData.pointsDeTache > 0 ) {
        let myroll = new Roll("2d10");
        myroll.roll();
        rollData.degats = parseInt(myroll.result) + parseInt(rollData.arme.data.dommages) + parseInt(this.data.data.attributs.plusdom.value);
        rollData.loc = RdDUtility.getLocalisation();
        for (let target of game.user.targets) {
          defenseMsg = RdDutility.buildDefenseChatCard(this, target, rollData );
          specialStr = "<br><strong>Cible</strong> : " + target.actor.data.name;
        }
        specialStr += "<br>Dommages : " + rollData.degats + "<br>Localisation : " + rollData.loc.label;
      } else {
        specialStr = "<br>Echec ! Pas de dommages";        
      }
    }

    // Save it for fight
    await this.setFlag( "foundryvtt-reve-de-dragon", "rollData", undefined );
    await this.setFlag( "foundryvtt-reve-de-dragon", "rollData", rollData );
    
    let chatOptions = { content:  "<strong>Test : " + rollData.selectedCarac.label + " / " + rollData.competence.name + "</strong><br>Jet : " + 
                                    rollData.selectedCarac.value + " / " + rollData.finalLevelStr + " - " + rollData.rollTarget.score + "%<br><strong>Résutat : </strong>" +  myroll.total + "<br>" +
                                    "<strong>" + quality + "</strong>" + specialStr + xpmsg,
                        user: game.user._id,
                        title:  "Résultat du test"
                      }
    ChatMessage.create( chatOptions );
    
    if ( defenseMsg ) { // target hit !    
      ChatMessage.create( defenseMsg );        
    } 
  }
  
  /* -------------------------------------------- */  
  updateCarac( caracName, caracValue )  
  {
    //let data = this.data.data;
    let caracpath = "data.carac." + caracName + ".value"
    this.update( { caracpath: caracValue } );
    //data.carac[caracName].value = caracValue; // Force update ?
    //RdDUtility.computeCarac( data );
  }
  
  /* -------------------------------------------- */  
  async updateCompetence( compName, compValue )
  {
    let comp = RdDUtility.findCompetence( this.data.items, compName);
    if ( comp ) {
      //console.log("ACTOR", this);
      //comp.data.niveau = compValue;
      const update = {_id: comp._id, 'data.niveau': compValue };
      const updated = await this.updateEmbeddedEntity("OwnedItem", update); // Updates one EmbeddedEntity
      //console.log("UP", updated);
    } else {
      console.log("Competence not found", compName);
    }
  }
  
  /* -------------------------------------------- */  
  computeEtatGeneral( ) 
  {
    let data = this.data.data;
    let state = 0;
    state = state - (data.sante.vie.max - data.sante.vie.value);
    state = state + RdDUtility.currentFatigueMalus(data.sante.fatigue.value, data.sante.endurance.max);
    data.compteurs.etat.value = state;
  }
  
  /* -------------------------------------------- */  
  testSiSonne( sante, endurance ) 
  {
    let myroll = new Roll("d20");
    myroll.roll();
    let result =  myroll.total;
    if ( result <= endurance.value)
       sante.sonne.value = false;
    if ( result > endurance.value || result == 20) // 20 is always a failure
       sante.sonne.value = true;
    if (result == 1) {
      sante.sonne.value = false;
      let xp  = parseInt(this.data.data.carac.constitution.xp) + parseInt(1);
      this.update( {"data.carac.constitution.xp": xp } ); // +1 XP !
      // TODO : Output to chat
    }
  }
  
  /* -------------------------------------------- */  
  async santeIncDec(name, inc ) {
    const sante = duplicate(this.data.data.sante);
    let data = sante[name];
    let lastValue = data.value; // Useful for Endurance and Sonné
    data.value = data.value + inc;
    if ( data.value > data.max ) data.value = data.max;
    if ( data.value < 0 ) data.value = 0;
    
    if (name == "endurance") {
      if ( inc < 0 ) // Each endurance lost -> fatigue lost
        sante.fatigue.value = sante.fatigue.value - inc
      
      // If endurance is 0 -> -1 vie
      if ( data.value == 0 )  {
        sante.vie.value = sante.vie.value - 1;
      }
      let diffVie = sante.vie.max - sante.vie.value;
      if ( data.value > data.max - (diffVie*2) ) {
        data.value = data.max - (diffVie*2);
      }
      
      let blessures = this.data.data.blessures;
      let maxEnd = Math.floor( data.max / blessures.graves.nombre);
      if (data.value > maxEnd ) data.value = maxEnd;
      if ( blessures.critiques.nombre > 0 && data.value > 1) data.value = 1;
      
      if (lastValue - data.value > 1) this.testSiSonne(sante, data); // Peut-être sonné si 2 points d'endurance perdus d'un coup
    }
    
    let diffEndurance = sante.endurance.max - this.data.data.sante.endurance.value;
    if ( sante.fatigue.value < diffEndurance) // If endurance lost, then the same amount of fatigue cannot be recovered
      sante.fatigue.value = diffEndurance;
    
    await this.update( {"data.sante": sante } );
  }
  
  /* -------------------------------------------- */  
  manageBlessures( blessuresData  ) 
  {
    let blessures = duplicate(this.data.data.blessures);
    if (blessuresData.legeres > 0 ) {
      blessures.legeres.nombre += 1;
      if ( blessures.legeres.nombre > 5 ) { 
        blessures.legeres.nombre = 5;
        blessuresData.graves += 1;
      } else {
        blessures.legeres.liste[blessures.legeres.nombre-1].etat = "active";
        blessures.legeres.liste[blessures.legeres.nombre-1].loc  = blessuresData.locName;
      } 
    }
    if ( blessuresData.graves > 0 ) { 
      blessures.graves.nombre += 1;
      if ( blessures.graves.nombre > 2 ) { 
        blessures.graves.nombre = 2;
        blessuresData.critiques += 1;
      } else {
        blessures.graves.liste[blessures.graves.nombre-1].etat = "active";
        blessures.graves.liste[blessures.graves.nombre-1].loc  = blessuresData.locName;
      } 
    }
    if ( blessuresData.critiques > 0 ) {
      blessuresData.endurance = this.data.data.sante.endurance.value; // Patch with real endurance current value (ie end -> 0 when critique)
      blessures.critiques.nombre = 1;
      blessures.critiques.liste[0].etat = "active";
      blessures.critiques.liste[0].loc  = blessuresData.locName;
    }
    if ( blessuresData.legeres > 0 || blessuresData.graves > 0 || blessuresData.critiques > 0 ) 
      this.update( { "data.blessures": blessures } );
  }
  
  /* -------------------------------------------- */  
  rollArme( armeName ) 
  {   
    let armeItem = RdDUtility.findCompetence( this.data.items, armeName );
    if ( armeItem && armeItem.data.competence ) 
      this.rollCompetence( armeItem.data.competence, armeItem );
    else 
      this.rollCompetence( armeName ); //Bypass mode!
  }
  
  /* -------------------------------------------- */  
  async rollCompetence( compName, armeItem ) 
  {
    console.log("!!!!!!", compName, armeItem);
    let compItem = RdDUtility.findCompetence( this.data.items, compName);
    let rollData = { 
        "competence": compItem, 
        "arme": armeItem,
        "carac": this.data.data.carac,
        "bonusmalusTable": CONFIG.RDD.bonusmalus,
        "etat": this.data.data.compteurs.etat.value, 
        "bmValue": 0,
        "finalLevel": 0
      }

    let html = await renderTemplate('systems/foundryvtt-reve-de-dragon/templates/dialog-competence.html', rollData);
    if (armeItem)  {
      new RdDRollDialog("arme", html, rollData, this ).render(true);
    } else {
      new RdDRollDialog("competence", html, rollData, this ).render(true);
    }
  }
  
  /* -------------------------------------------- */  
  computeArmure( locData ) 
  {
    /* TODO */
    return 0;
  }
  
  /* -------------------------------------------- */  
  encaisserDommages( attackerActor ) 
  {
    let attackerRoll = attackerActor.getFlag("foundryvtt-reve-de-dragon", "rollData");
    console.log("RollData!", attackerActor, attackerRoll);
    let degatsReel   = attackerRoll.degats - this.computeArmure(attackerRoll.loc);
    let result = RdDUtility.computeBlessuresSante(degatsReel);
    this.santeIncDec("vie", result.vie);
    this.santeIncDec("endurance", result.vie);
    
    result.locName = attackerRoll.loc.label; // Add the localisation namme
    this.manageBlessures( result ); // Will upate the result table
    ChatMessage.create( {title: "Blessures !", content: this.data.name + " a encaissé : " +
      "<br>" + result.legeres    + " légères, " + result.graves     + " graves et " +
      result.critiques  + " critique." +
      "<br>Et perdu : " + 
      "<br>" + result.endurance + " Endurance et " + result.vie + " Points de Vie" } );
  
    this.computeEtatGeneral();
    this.sheet.render(true);
  }
  
  /* -------------------------------------------- */  
  parerAttaque( attackerActor, armeId )
  {
    console.log("Going to PARY !!!!!!!!!");
  }
  
  /* -------------------------------------------- */  
  /** @override */
  getRollData() {
    const data = super.getRollData();
    const shorthand = game.settings.get("foundryvtt-reve-de-dragon", "macroShorthand");

    // Re-map all attributes onto the base roll data
    if ( !!shorthand ) {
      for ( let [k, v] of Object.entries(data.attributes) ) {
        if ( !(k in data) ) data[k] = v.value;
      }
      delete data.attributes;
    }

    // Map all items data using their slugified names
    data.items = this.data.items.reduce((obj, i) => {
      let key = i.name.slugify({strict: true});
      let itemData = duplicate(i.data);
      if ( !!shorthand ) {
        for ( let [k, v] of Object.entries(itemData.attributes) ) {
          if ( !(k in itemData) ) itemData[k] = v.value;
        }
        delete itemData["attributes"];
      }
      obj[key] = itemData;
      return obj;
    }, {});
    return data;
  }
}
